/*
 * Mach64.c - Mach64 device driver for the Linux GGI
 * Copyright (C) 1997	Jason McMullan <jmcc@cs.cmu.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or
 * Jason McMullan, 5317 WEH, Carnegie Mellon University, 5000 Forbes Ave,
 * Pittsburgh, PA 15213. The author can be contacted via email at 
 * jmcc@cs.cmu.edu.
 *
 * $Log: Mach64.c,v $
 * Revision 1.3  1997/12/09 08:20:22  seeger
 * fixed to new input scheme.
 *
 * Revision 1.2  1997/04/23 03:52:43  ezrec
 * Minor: Working on the ATI Mach64...
 *
 * Revision 1.1	1997/03/21 23:25:54	ezrec
 * Major: ATI Mach64 driver. Doesn't work. Tries to detect the Mach64 card.
 *				PCI only. Have fun!
 *
 * Revision 1.2	1997/01/20 04:15:59	ezrec
 * Minor: Now finds the frame buffer - still needs work... And it does look
 *	 like I'll have to dig up the $50 for the ATI technical info - the
 *	 MMIO registers don't seem to cover text mode (font setting, etc..)
 *	 If at all possible, I'd like to use _only_ the MMIO regs, to stay
 *	 away from the VGA nastyness.....
 *
 * Revision 1.1	1997/01/18 22:12:59	ezrec
 * Major:	Adding support for the ATI MAch64 chipset, currently only
 *	detects the Mach64 (PCI only!)
 *
 */
#define DEBUG 1
#define MAINTAINER "Jason McMullan <jmcc@cs.cmu.edu>"

static char *rcsid="$Id: Mach64.c,v 1.3 1997/12/09 08:20:22 seeger Exp $";

/* Configuration options */

/* GGI includes */
#include <ggi/kgi.h>
#include <ggi/module.h>

/* Kernel includes */
#include <asm/dma.h>		/* stuff for DMA transfers */
#include <asm/io.h>		/* I/O routines (inb and outb macros) */
#include <asm/system.h>
#include <linux/delay.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/bios32.h>	/* For pcibios_* routines */
#include <linux/pci.h>
#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/ioport.h>
#include <linux/time.h>
#include <linux/timer.h>
#include <linux/stat.h>

#include "ATI/Mach64.h"

#define DPRINTK(x, a...) TRACE(printk(KERN_DEBUG x, ##a))

#define MAX_NR_ATI 4

#define DEF_INDEX 0

typedef uint32 ati_reg;

static struct ati_config_s {
	 uint8 pci_bus;		/* Which bus the device is on? */
	 uint8 pci_devfn;	/* The device ID? */
	 uint32	rev; 	 	/* ATI revision code */

	 struct kgi_chipset chipinfo;	/* Chipset info */
	 ati_reg default[NUM_CTL_REGS];	/* Default registers */

	 uint8 old_CONFIG_CNTL;
	 uint8 old_MEM_CNTL;
	 uint8 ioCONFIG_CNTL;
	 uint8 ioMEM_CNTL;

	 /* Address space mappings. */
	 struct io_region io;		/* Regular ISA Io Region */
	 struct mem_region mmio;	/* MMIO Region (FB + Regs) */

	 ati_reg *reg_map;		/* Pointer to MMIO registers */
} ati[MAX_NR_ATI];

static int ati_index=0;

#include "Mach64_io.inc"
#include "Mach64_mapping.inc"
#include "Mach64_strings.inc"

/* kgi_chipset_init
 *
 * This subroutine handles device initialization.	It returns EOK on
 * success and -ECHIPSET on failure (or error).
 */
int kgi_chipset_init (void)
{
	DPRINTK("kgi_chipset_init: Mach64 (PCI) driver version $Version$.\n");

	if (ati_index >= MAX_NR_ATI ) {
		printk(KERN_ERR"No more Mach64 PCI cards can be registered\n");
		return -E(CHIPSET,BOGUS);
	}

	if (Mach64_detect(ati_index)==EOK &&
	    Mach64_init(ati_index)==EOK) {
		ati_index++;
		return EOK;
	}

	return -E(CHIPSET,BOGUS);

}


/* kgi_chipset_done
 * This subroutine shuts the driver down.
 */
void kgi_chipset_done (void)
{
	while (ati_index) {
		Mach64_release(ati_index);
		ati_index--;
	}
	DPRINTK("kgi_chipset_done: Removed driver.\n");
}

struct kgi_chipset *kgi_chipset_get_info (void)
{
	return &ati[DEF_INDEX].chipinfo; /* this should suffice */
}

int kgi_chipset_check_mode(struct kgi_mode *mode,enum kgi_tm_cmd cmd)
{
	unsigned int i,reg1,reg2;
	int rlist[]={
		sioCONFIG_CHIP_ID,
		sioCONFIG_CNTL,
		sioSCRATCH_REG0,
		sioSCRATCH_REG1,
		sioCONFIG_STAT0,
		sioMEM_CNTL,
		sioGEN_TEST_CNTL,
		sioCLOCK_CNTL,
		sioCRTC_GEN_CNTL};
	int mlist[]={
		CONFIG_CHIP_ID,
		CONFIG_CNTL,
		SCRATCH_REG0,
		SCRATCH_REG1,
		CONFIG_STAT0,
		MEM_CNTL,
		GEN_TEST_CNTL,
		CLOCK_CNTL,
		CRTC_GEN_CNTL};
	char *slist[]={
		"CONFIG_CHIP_ID",
		"CONFIG_CNTL",
		"SCRATCH_REG0",
		"SCRATCH_REG1",
		"CONFIG_STAT0",
		"MEM_CNTL",
		"GEN_TEST_CNTL",
		"CLOCK_CNTL",
		"CRTC_GEN_CNTL"};

	for (i=0;i<9;i++) {
		reg1=REGr(mlist[i]);
		if (ioBASE!=0) {
			reg2=inl(ioBASE|(rlist[i]<<10));
		} else {
			reg2=reg1;
		}
		DPRINTK("[%-14s] sio(0x%.8x) mmio(0x%.8x)\n",slist[i],reg2,reg1);
	}
	return -E(CHIPSET,BOGUS);
}

void kgi_chipset_set_mode(struct kgi_mode *mode)
{

}

int kgi_chipset_ioctl(struct inode *inode, struct file *file, 
											unsigned int cmd,unsigned long arg)
{
	return -E(CHIPSET,BOGUS);
}

int kgi_graphics_ioctl(struct inode *inode, struct file *file, 
											 unsigned int cmd,unsigned long arg)
{
	return -E(CHIPSET,BOGUS);
}
