/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>

#ifdef _AIX
int setitimer (int Which, const struct itimerval *value, struct itimerval *ovalue);
#endif

#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <signal.h>
#include <string.h>

#include "ggi-dl.h"
#include "internal.h"
#include "Xvisual.h"

#include <ggi/ggi_commands.h>

extern int        _GGIbpc;
extern int        _GGInocols;
extern ggi_color *_GGIcolormap;

int GGIsetfocus(ggi_visual *vis)
{
	return EOK;
}

int GGIsetorigin(ggi_visual *vis,int x,int y)
{
	struct Xhooks *xhook=VIS_PRIVATE(vis);
	ggi_mode *mode=VIS_MODE(vis);

	if ( x<0 || x> mode->virt.x-mode->visible.x ||
	     y<0 || y> mode->virt.y-xhook->ysplit )
	     return -EINVAL;

	xhook->xoff=x;
	xhook->yoff=y;
	ggiFlush(vis);
	
	return EOK;
}

int GGIsetsplitline(ggi_visual *vis,int y)
{
	struct Xhooks *xhook;
        if (y<0 || y>VIS_MODE(vis)->visible.y) return -1;
        
	xhook=VIS_PRIVATE(vis);

	if ( xhook->yoff > VIS_MODE(vis)->virt.y-y )
	     xhook->yoff = VIS_MODE(vis)->virt.y-y;

	xhook->ysplit=y;
	ggiFlush(vis);
	
	return EOK;
}

static int _GGIdomode(ggi_visual *vis)
{
	int err;
	ggi_suggest sug;

	char sugs[1][80];
	
	switch(VIS_MODE(vis)->graphtype)
	{
	  case GT_8BIT : err=8;break;
	  case GT_15BIT: err=15;break;
	  case GT_16BIT: err=16;break;
	  case GT_24BIT: err=24;break;
	  case GT_32BIT: err=32;break;
	  default: return -EINVAL;
	}

	sprintf(sugs[0],"generic-linear-%d",err);

	_ggiZapMode(vis,~GGI_DL_OPDISPLAY);

	_ggiSetupMode(vis);

	err=(_ggiOpenDL(vis,"generic-stubs","")==NULL);
	if (err) {
		fprintf(stderr,"display-X: Can't open the \"generic-stubs\" "
				"library.\n");
	}

	sug.id=0;
	do {
		/* FIXME ! */
		strcpy(sug.name,sugs[sug.id]);
		strcpy(sug.args,"");

		err=(_ggiOpenDL(vis,sug.name,sug.args)==NULL);
		if (err) {
			fprintf(stderr,"display-X: Can't find an appropriate "
				       "library for %s (%s)\n",
					sug.name,sug.args);
			return err;
		} else {
			DPRINT("Success in loading %s (%s)\n",sug.name,sug.args);
		}
	} while (sug.id!=0);

	/* Override the color operations for X if we are 8 bit palette. */
	if (_GGIcolormap) { 
		free(_GGIcolormap);_GGIcolormap=NULL;
	}

	if (VIS_MODE(vis)->graphtype==GT_8BIT) {

		int x;
	        _GGIbpc=1;
		_GGInocols=256;
		_GGIcolormap=malloc(256*sizeof(ggi_color));
		for(x=0;x<_GGInocols;x++)
		{
			_GGIcolormap[x].r=((x&0xe0)>>5)*0x2492;
			_GGIcolormap[x].g=((x&0x1c)>>2)*0x2492;
			_GGIcolormap[x].b=((x&0x03)   )*0x5555;
		} 
		vis->opcolor->mapcolor=GGImapcolor;
		vis->opcolor->unmappixel=GGIunmappixel;
		vis->opcolor->setpalvec=GGIsetpalvec;
		vis->opcolor->getpalvec=GGIgetpalvec;
		/* This is just to trigger the initial setup. */
		GGIsetpalvec(vis,0,256,_GGIcolormap);
	}

	/* X can splitline and pan on a larger bitmap ;-) */
        vis->op2d->setsplitline=GGIsetsplitline;
	vis->op2d->setorigin=GGIsetorigin;
        
	return EOK;
}

static int GGIdoflush(struct Xhooks *xhook)
{
	
	if (xhook->XLibLock++!=0) 
		{xhook->XLibLock--;return -EAGAIN;}	/* FIXME use spinlocks here !*/

	if (xhook->cmap && xhook->cmap_first<xhook->cmap_last) {

		int x;
		XColor xcol;
	       	for(x=xhook->cmap_first;x<xhook->cmap_last;x++)
       		{
       			xcol.red  =_GGIcolormap[x].r;
	       		xcol.green=_GGIcolormap[x].g;
       			xcol.blue =_GGIcolormap[x].b;
       			xcol.pixel=x;
	       		xcol.flags= DoRed | DoGreen | DoBlue ;
       			XStoreColor(xhook->display, xhook->cmap,&xcol);
	       	}
		xhook->cmap_first=256;
		xhook->cmap_last=0;
		XSetWindowColormap(xhook->display, xhook->window, xhook->cmap);
	}

	if (xhook->have_shm)
	{
		/* It seems X cannot pan horizontally, when using XShm. */
		XShmPutImage(xhook->display, xhook->window, xhook->gc,
			  xhook->ximage, 
			  /* xhook->xoff */ 0, xhook->yoff, /* panning ! */
			  0, 0,
			  xhook->ximage->width, 
			  xhook->ysplit, 0);

		if (xhook->ysplit<xhook->ximage->height)
			XShmPutImage(xhook->display, xhook->window, xhook->gc,
				  xhook->ximage, 
				  0, 0,
				  0, xhook->ysplit,
				  xhook->ximage->width, 
				  xhook->ximage->height-xhook->ysplit, 0);

		XSync(xhook->display,0);
	} else {

		XPutImage(xhook->display, xhook->window, xhook->gc,
			  xhook->ximage, 
			  xhook->xoff, xhook->yoff, /* panning ! */
			  0, 0,
			  xhook->ximage->width, 
			  xhook->ysplit);

		if (xhook->ysplit<xhook->ximage->height)
			XPutImage(xhook->display, xhook->window, xhook->gc,
				  xhook->ximage, 
				  0, 0,
				  0, xhook->ysplit, /* panning ! */
				  xhook->ximage->width, 
				  xhook->ximage->height-xhook->ysplit);

		XFlush(xhook->display);
		XSync(xhook->display,0);
	}

	xhook->XLibLock--;	/* FIXME Spinlocks ... */

	return EOK;
}

static struct Xhooks *xhk;

static void _doflush_ALRM(int unused)
{ 
	struct itimerval tval={{0,00000},{0,50000}};
	signal(ASYNCSIGNAL,_doflush_ALRM);
	GGIdoflush(xhk);
	setitimer(ASYNCTIMER,&tval,NULL);
}

int GGIflush(ggi_visual *vis)
{
	if (VIS_FLAGS(vis) & GGIFLAG_ASYNC) {
		struct itimerval tval={{0,00000},{0,0}};
		signal(ASYNCSIGNAL,_doflush_ALRM);
		setitimer(ASYNCTIMER,&tval,NULL);
	} else {
		struct itimerval tval={{0,00000},{0,50000}};
		signal(ASYNCSIGNAL,_doflush_ALRM);
		setitimer(ASYNCTIMER,&tval,NULL);
	}
	return GGIdoflush((struct Xhooks *)VIS_PRIVATE(vis));
}

/* Hack to fall back when shm is not working. */
static int	shmerror;
static int	(*oldshmerrorhandler) (Display *, XErrorEvent *);

static int shmerrorhandler (Display * disp, XErrorEvent * event)
{
	if (event->error_code == BadAccess)
		shmerror=1;
	else
		(*oldshmerrorhandler) (disp, event);
	return 0;
}

static int _GGIbasiccheck(ggi_visual *vis,ggi_mode *tm,	XVisualInfo *vinfo)
{
	struct Xhooks *xhook=VIS_PRIVATE(vis);

	if (vis==NULL) {
		DPRINT("Visual==NULL\n");
		return -1;
	}
	
	switch(tm->graphtype) { 
	    case GT_8BIT : 
	    	tm->virt.x=((tm->virt.x+3) & ~3);
		if ( ! XMatchVisualInfo (xhook->display, xhook->screen, 8, PseudoColor, vinfo) )
		{
			DPRINT("No 8 Bit PseudoColor Visual found. Sorry.\n");
			return GGI_DL_ERROR;
		}
		break;
	    case GT_15BIT: 
		tm->virt.x=((tm->virt.x+1) & ~1);
		if ( ! XMatchVisualInfo (xhook->display, xhook->screen, 15, TrueColor, vinfo) )
		{
			DPRINT("No 15 Bit TrueColor Visual found. Sorry.\n");
			return GGI_DL_ERROR;
		}
		break;
	    case GT_16BIT: 
	    	tm->virt.x=((tm->virt.x+1) & ~1);
		if ( ! XMatchVisualInfo (xhook->display, xhook->screen, 16, TrueColor, vinfo) )
		{
			DPRINT("No 16 Bit TrueColor Visual found. Sorry.\n");
			return GGI_DL_ERROR;
		}
		break;
	    case GT_24BIT:
		if ( ! XMatchVisualInfo (xhook->display, xhook->screen, 24, TrueColor, vinfo) )
		{
			DPRINT("No 24 Bit TrueColor Visual found. Sorry.\n");
			return GGI_DL_ERROR;
		}
		break;
	    case GT_32BIT: 
		if ( ! XMatchVisualInfo (xhook->display, xhook->screen, 24, TrueColor, vinfo) )
		{
			DPRINT("No 24 Bit TrueColor Visual for 32 Bit found. Sorry.\n");
			return GGI_DL_ERROR;
		}
		break;
	    default:
		DPRINT("Cannot use this graphics type with display-X. Sorry.\n");
	  	return GGI_DL_ERROR;
	}

	return 0;
}

int GGIsetmode(ggi_visual *vis,ggi_mode *tm)
{ 
	struct Xhooks *xhook=VIS_PRIVATE(vis);
	XSizeHints hint;
	unsigned int fg, bg;
	XVisualInfo vinfo;
	int rc;
	
	if ((rc=_GGIbasiccheck(vis,tm,&vinfo))) return rc;

	xhook->xoff=xhook->yoff=0;
	xhook->ysplit=tm->visible.y;

        signal(ASYNCSIGNAL, SIG_IGN);	/* We disable the updating thread here. Thus no locking necessary. */

	/* Fill in hint structure. Minimal window until we know better. */
	hint.x      = hint.y      = 0;
	hint.width  = tm->visible.x;
	hint.height = tm->visible.y;
	hint.flags  = PSize;
  
	/* Get some colors */
	bg = WhitePixel (xhook->display, xhook->screen);
	fg = BlackPixel (xhook->display, xhook->screen);
	DPRINT("X-lib has colors.\n");
  
	if (xhook->window) 
		XDestroyWindow(xhook->display,xhook->window);

	if (xhook->ximage)
		XDestroyImage(xhook->ximage);

	if (xhook->shminfo.shmid!=-1) {	
		XShmDetach(xhook->display, &(xhook->shminfo));
		DPRINT("Detach\n");
		if (VIS_FB_LINEAR(vis)!=NULL) {
			shmdt(VIS_FB_LINEAR(vis));
		}
	}

	if (VIS_FB_LINEAR(vis)!=NULL) {
		VIS_FB_LINEAR(vis)=NULL;
		VIS_FB_LINEAR_SIZE(vis)=0;
	}


	xhook->window = XCreateSimpleWindow (xhook->display,
				 RootWindow (xhook->display,xhook->screen),
				 hint.x, hint.y,
				 hint.width, hint.height,
				 2, fg, bg);
  
	DPRINT("X-lib has window.\n");

	XSelectInput(xhook->display, xhook->window, 
	             KeyPressMask|KeyReleaseMask|
	             ButtonPressMask|ButtonReleaseMask|
	             PointerMotionMask
	             );

	/* Tell other applications about this window */
	XSetStandardProperties (xhook->display, xhook->window, "GGI-on-X", "GGI-on-X", None, NULL, 0, &hint);

	if (!xhook->gc)
	  	xhook->gc = XCreateGC(xhook->display, xhook->window, 0, 0);
	DPRINT("X-lib GC allocated.\n");

	if (xhook->have_shm)
	{
		DPRINT("X: mode setting\n");
		xhook->ximage = XShmCreateImage(
			xhook->display, vinfo.visual, vinfo.depth,
			ZPixmap, 
			0, 
			&(xhook->shminfo),
			tm->virt.x,tm->virt.y);

#ifdef LITTLE_ENDIAN_ARCHITECTURE
		xhook->ximage->byte_order = LSBFirst;
		xhook->ximage->bitmap_bit_order = LSBFirst;
#else
		xhook->ximage->byte_order = MSBFirst;
		xhook->ximage->bitmap_bit_order = MSBFirst;
#endif

		xhook->shminfo.shmid = 
			shmget(IPC_PRIVATE, (unsigned int) 
				tm->virt.x*tm->virt.y*vinfo.depth,
				IPC_CREAT | 0777);

		VIS_FB_LINEAR(vis)=xhook->shminfo.shmaddr=xhook->ximage->data = 
			(char *) shmat(xhook->shminfo.shmid,0,0);
		xhook->shminfo.readOnly=False;
	
		DPRINT("X: shmat success at %p.\n",VIS_FB_LINEAR(vis));

		/* Argh ... X sometimes says it can XShm but it can't ... */
		shmerror = 0;
		oldshmerrorhandler = XSetErrorHandler(shmerrorhandler);
		XShmAttach(xhook->display, &(xhook->shminfo));

		XSync(xhook->display,0);
		XSetErrorHandler(oldshmerrorhandler);

		if (shmerror) {
			if (VIS_FB_LINEAR(vis)!=NULL) {
				shmdt(VIS_FB_LINEAR(vis));
				VIS_FB_LINEAR(vis)=NULL;
			}
			if (xhook->ximage) {
				XDestroyImage(xhook->ximage);
				xhook->ximage=NULL;
			}
			xhook->shminfo.shmid = -1;
			xhook->have_shm = 0;
			DPRINT("X: XShmAttach failed - trying without shm.\n");
		} else {
			shmctl(xhook->shminfo.shmid, IPC_RMID, 0);
			DPRINT("X: ShmAllDone.\n");
		}
	}

	if (! xhook->have_shm) {

		VIS_FB_LINEAR(vis)= 
			(char *) malloc(tm->virt.x*tm->virt.y*vinfo.depth);
		DPRINT("X: Buffer allocated %p.\n",VIS_FB_LINEAR(vis));

		xhook->ximage = XCreateImage(
			xhook->display, vinfo.visual, vinfo.depth,
			ZPixmap, /* format */
			0, 	 /* offset */
			VIS_FB_LINEAR(vis),	/* data */
			tm->virt.x,tm->virt.y, /* size */
			8,	/* bitmap_pad*/
			0 
			);

		DPRINT("X: Image allocated.\n");
	}

	XResizeWindow(xhook->display, xhook->window, tm->visible.x, tm->visible.y);
	DPRINT("X: Resize.\n");

	DPRINT("About to map.\n");
	/* Map window. */
	XMapRaised(xhook->display, xhook->window);

	if (xhook->cmap)
		XFreeColormap(xhook->display,xhook->cmap);

	if (tm->graphtype==GT_8BIT)
	{
	       	xhook->cmap = XCreateColormap(xhook->display, xhook->window, 
       				vinfo.visual, 
       				AllocAll);

		xhook->cmap_first=256;
		xhook->cmap_last=0;
		                                
		DPRINT("X-lib colormap allocated %x.\n",xhook->cmap);
	} else {
		xhook->cmap=0;
	}
	

	XFlush(xhook->display);
	DPRINT("X: Flush.\n");

	GGIdoflush(xhk=xhook);
	signal(ASYNCSIGNAL,_doflush_ALRM);
	raise(ASYNCSIGNAL);

	memcpy(VIS_MODE(vis),tm,sizeof(ggi_mode));

	return _GGIdomode(vis);
}

/**********************************/
/* check any mode (text/graphics) */
/**********************************/
int GGIcheckmode(ggi_visual *vis,ggi_mode *tm)
{
	XVisualInfo vinfo;	/* Dummy here. setmode needs it ... */
	int rc;

	if ((rc=_GGIbasiccheck(vis,tm,&vinfo))) return rc;	
	return 0;
}

/************************/
/* get the current mode */
/************************/
int GGIgetmode(ggi_visual *vis,ggi_mode *tm)
{
	DPRINT("In GGIgetmode(%p,%p)\n",vis,tm);
	if (vis==NULL)
		return -1;
	
	/* We assume the mode in the visual to be o.k. */
	memcpy(tm,VIS_MODE(vis),sizeof(ggi_mode));

	return 0;
}
