/* --------------------------------------------------------------------------
**	i386 PC keyboard driver
** --------------------------------------------------------------------------
**
**	Copyright (C) 1996-1997 by Steffen Seeger
**
**	NOTE: Additional copyrights apply as this code was derived from
**	the linux keyboard handler. You have to apply these yourself because
**	for using it with GGI, you need the old source anyway.
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Steffen_Seeger

#include <ggi/kgi.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/random.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <asm/io.h>

#define	KEYBOARD_IRQ	1

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,37)
# define test_and_clear_bit(x,y) clear_bit(x,y)
# define test_and_set_bit(x,y) set_bit(x,y)
#endif

/*
**	register definitions
*/

/*	keyboard controller registers
*/
#define KB_STATUS_REG	0x64
#define KB_CNTL_REG	0x64
#define KB_DATA_REG	0x60

/*	controller commands
*/
#define KB_READ_MODE	0x20
#define KB_WRITE_MODE	0x60
#define KB_SELF_TEST	0xAA
#define KB_SELF_TEST2	0xAB
#define KB_CNTL_ENABLE	0xAE

/*	keyboard commands
*/
#define KB_ENABLE	0xF4
#define KB_DISABLE	0xF5
#define KB_RESET	0xFF

/*	keyboard replies
*/
#define KB_ACK		0xFA
#define KB_POR		0xAA

/*	status register bits
*/
#define KB_OBF		0x01
#define KB_IBF		0x02
#define	KB_ODS		0x20
#define KB_GTO		0x40
#define KB_PERR		0x80

/*	keyboard controller mode register bits
*/
#define KB_EKI		0x01
#define KB_SYS		0x04
#define KB_DMS		0x20
#define KB_KCC		0x40

#define TIMEOUT_CONST	0x100000

uchar psaux_read_mask = 0x01;	/* modified by psaux.c		*/

#define	MAX_NR_KEYS		128	/* maximum 256 */
#define	MAX_NR_MODIFIERS	8

struct kbd_struct {

	struct kgi_input input;

	ggi_uint	resend;
	ggi_uint	reply_expected;
	ggi_uint	acknowledge;
	ggi_uint	repeat;
	ggi_uint	leds; 
	ggi_uint	prev_scancode;

	#define	BYTES(x)	(((x) + 7) & ~7)
	uchar	down_keys[BYTES(MAX_NR_KEYS)];
};

/*	scancode to keycode translation
**
**	This is now user-settable. The keycodes 1-88,96-111,119 are fairly
**	standard, and should probably not be changed - changing might confuse X.
**	X also interprets scancode 0x5d (KEY_Begin). For 1-88 keycode equals
**	scancode.
*/

#define E0_KPENTER 96
#define E0_RCTRL   97
#define E0_KPSLASH 98
#define E0_PRSCR   99
#define E0_RALT    100
#define E0_BREAK   101  /* (control-pause) */
#define E0_HOME    102
#define E0_UP      103
#define E0_PGUP    104
#define E0_LEFT    105
#define E0_RIGHT   106
#define E0_END     107
#define E0_DOWN    108
#define E0_PGDN    109
#define E0_INS     110
#define E0_DEL     111

#define E1_PAUSE   119

/*	The keycodes below are randomly located in 89-95,112-118,120-127.
**	They could be thrown away (and all occurrences below replaced by 0),
**	but that would force many users to use the `setkeycodes' utility, where
**	they needed not before. It does not matter that there are duplicates, as
**	long as no duplication occurs for any single keyboard.
*/

#define SC_LIM 89

#define FOCUS_PF1 85           /* actual code! */
#define FOCUS_PF2 89
#define FOCUS_PF3 90
#define FOCUS_PF4 91
#define FOCUS_PF5 92
#define FOCUS_PF6 93
#define FOCUS_PF7 94
#define FOCUS_PF8 95
#define FOCUS_PF9 120
#define FOCUS_PF10 121
#define FOCUS_PF11 122
#define FOCUS_PF12 123

#define JAP_86     124

/*	tfj@olivia.ping.dk: The four keys are located over the numeric 
**	keypad, and are labelled A1-A4. It's an rc930 keyboard, from
**	Regnecentralen/RC International, Now ICL. Scancodes: 59, 5a, 5b, 5c.
*/

#define RGN1 124
#define RGN2 125
#define RGN3 126
#define RGN4 127

static uint8 high_keys[128 - SC_LIM] = {

	RGN1, RGN2, RGN3, RGN4, 0, 0, 0,                   /* 0x59-0x5f */
	0, 0, 0, 0, 0, 0, 0, 0,                            /* 0x60-0x67 */
	0, 0, 0, 0, 0, FOCUS_PF11, 0, FOCUS_PF12,          /* 0x68-0x6f */
	0, 0, 0, FOCUS_PF2, FOCUS_PF9, 0, 0, FOCUS_PF3,    /* 0x70-0x77 */
	FOCUS_PF4, FOCUS_PF5, FOCUS_PF6, FOCUS_PF7,        /* 0x78-0x7b */
	FOCUS_PF8, JAP_86, FOCUS_PF10, 0                   /* 0x7c-0x7f */
};

/* BTC */
#define E0_MACRO   112
/* LK450 */
#define E0_F13     113
#define E0_F14     114
#define E0_HELP    115
#define E0_DO      116
#define E0_F17     117
#define E0_KPMINPLUS 118

/*	kkoller@nyx10.cs.du.edu: My OmniKey generates e0 4c for the "OMNI" 
**	key and the right alt key does nada.
*/

#define E0_OK	124

/*	New microsoft keyboard is rumoured to have
**	e0 5b (left window button),	LBANNER or Windows_L
**	e0 5c (right window button),	RBANNER or Windows_R
**	e0 5d (menu button).		RMENU   or TaskMan
*/

#define E0_MSLW	125
#define E0_MSRW	126
#define E0_MSTM	127

static uint8 e0_keys[128] = {

	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x00-0x07 */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x08-0x0f */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x10-0x17 */
	0, 0, 0, 0, E0_KPENTER, E0_RCTRL, 0, 0,		/* 0x18-0x1f */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x20-0x27 */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x28-0x2f */
	0, 0, 0, 0, 0, E0_KPSLASH, 0, E0_PRSCR,		/* 0x30-0x37 */
	E0_RALT, 0, 0, 0, 0, E0_F13, E0_F14, E0_HELP,	/* 0x38-0x3f */
	E0_DO, E0_F17, 0, 0, 0, 0, E0_BREAK, E0_HOME,	/* 0x40-0x47 */
	E0_UP, E0_PGUP, 0, E0_LEFT, E0_OK, E0_RIGHT, E0_KPMINPLUS, E0_END,/* 0x48-0x4f */
	E0_DOWN, E0_PGDN, E0_INS, E0_DEL, 0, 0, 0, 0,	/* 0x50-0x57 */
	0, 0, 0, E0_MSLW, E0_MSRW, E0_MSTM, 0, 0,	/* 0x58-0x5f */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x60-0x67 */
	0, 0, 0, 0, 0, 0, 0, E0_MACRO,			/* 0x68-0x6f */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x70-0x77 */
	0, 0, 0, 0, 0, 0, 0, 0				/* 0x78-0x7f */
};

static int kbd_set_keycode(ggi_uint scancode, ggi_uint keycode)
{
	if ((scancode < SC_LIM) || (scancode > 255) || (keycode > 127)) {

		return -EINVAL;
	}

	if (scancode < 128) {

		high_keys[scancode - SC_LIM] = keycode;

	} else {

		e0_keys[scancode - 128] = keycode;
	}

	return EOK;
}

static int kbd_get_keycode(ggi_uint scancode)
{
	return ((scancode < SC_LIM) || (scancode > 255))
			? -EINVAL
			: (scancode < 128)
				? high_keys[scancode - SC_LIM]
				: e0_keys[scancode - 128];
}

static struct kbd_struct kbd = 
{
	{
		NULL,			/* next			*/
		0,			/* focus		*/
		0,			/* device		*/
		emKey | emRawData,	/* events generated	*/
		emKey,			/* events to report	*/

		"IBM",			/* manufacturer		*/
		"PC-AT keyboard",	/* model		*/
		NULL			/* control function	*/
	},
	/* init the rest to 0 */
};

/*
**	driver code
*/

static inline void kbd_wait(void)
{
	int i;

	for (i = 0; i < TIMEOUT_CONST; i++) {

		if ((io_in8(KB_STATUS_REG) & KB_IBF) == 0) {

			return;
		}
	}

	#ifdef KBD_REPORT_ERR
		printk(KERN_WARNING "keyboard timed out\n");
	#endif
}

static inline void kbd_send_cmd(uchar cmd)
{
	kbd_wait();
	io_out8(cmd, KB_CNTL_REG);
}

/*	kbd_send_date() sends data to the keyboard and waits for acknowledge,
**	possibly returning if asked to do so. Returns 1 if failed, 0 on success.
*/
static int kbd_send_data(uchar data)
{
	int retries = 3;
	int i;

	do {
	 	kbd_wait();
		kbd.acknowledge = 0;
		kbd.resend = 0;
		kbd.reply_expected = 1;
		io_out8(data, KB_DATA_REG);

		for (i = 0; i < 0x200000; i++) {

			io_in8(KB_STATUS_REG);	/* just as a delay	*/

			if (kbd.acknowledge) {

				return 1;
			}

			if (kbd.resend) {

				break;
			}
		}

		if (!kbd.resend) {

			return 0;
		}

	} while (retries--);

	return 0;
}

static void kbd_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	ggi_event event;		/* struct for event reporting	*/
	uint8  scancode, keycode, release, status;

	/*	disable keyboard
	*/
	kbd_send_cmd(0xAD);
	kbd_wait();

	status = io_in8(KB_STATUS_REG);

	if ((status & psaux_read_mask) != 0x01) {

		/* 	On some platforms (Alpha XL for one), the init code
		**	may leave an interrupt hanging, yet with status
		**	indicating no data. After making sure that there's
		**	no data indicated and it's not a mouse interrupt, we
		**	will read the data register to clear it. If we don't
		**	do this, the data reg stays full and will not allow
		**	new data or interrupt from the keyboard. Sigh...
		*/
		if (!(status & (KB_ODS | KB_OBF))) {

			/* neither ODS nor OBF */
			scancode = io_in8(KB_DATA_REG);
		}

		goto end_kbd_intr;
	}

	scancode = io_in8(KB_DATA_REG);
	add_keyboard_randomness(scancode);

	if (kbd.reply_expected) {

		/*	0xfa, 0xfe only mean "acknowledge", "resend" for most
		**	keyboards but they are the key-up scancodes for PF6,
		**	PF10 on a FOCUS 9000
		*/
		kbd.reply_expected = 0;
		if (scancode == KB_ACK) {

			kbd.acknowledge = 1;
			goto end_kbd_intr;
		}

		if (scancode == 0xfe) {

			kbd.resend = 1;
			goto end_kbd_intr;
		}

		kbd.reply_expected = 1;

		#ifdef KBD_REPORT_ERR	/* strange, we better report...	*/

			DEBUGMSG("reply expected - got %02x", scancode);
		#endif
	}

	if (scancode == 0) {

		#ifdef KBD_REPORT_ERR

			printk(KERN_INFO "%s: buffer overflow\n",__FILE__);
		#endif
		kbd.prev_scancode = 0;
		goto end_kbd_intr;
	}

	event.any.focus		= kbd.input.focus;
	event.any.device	= kbd.input.id;
	event.any.time		= jiffies;

	if (kbd.input.report & emRawData) {

		event.raw.type		= evRawData;
		event.raw.size		= sizeof(ggi_any_event) + 1;
		event.raw.data[0]	= scancode;

		kgi_handle_input(&event);
	}

	if (scancode == 0xff) {

		#if 0
		/* ???	Why is this only an error if not in raw mode?
		** ???	rawmode did not change the operation of the keyboard!?
		*/

		/*	In scancode mode 1, my ESC key generates 0xff.
		**	The calculator keys on a FOCUS 9000 generate 0xff
		*/
			#ifndef KBD_IS_FOCUS_9000
				#ifdef KBD_REPORT_ERR
				if (!raw_mode) {

					DEBUGMSG("keyboard error");
				}
				#endif
			#endif
		#endif

		kbd.prev_scancode = 0;
		goto end_kbd_intr;
	}

	if (scancode == 0xe0 || scancode == 0xe1) {

		kbd.prev_scancode = scancode;
		goto end_kbd_intr;
	}

	/*	Convert scancode to keycode, using prev_scancode.
	*/

	release = (scancode & 0x80);
	scancode &= 0x7f;

	if (kbd.prev_scancode) {

		/*	usually it will be 0xe0, but a Pause key generates
		**	e1 1d 45 e1 9d c5 when pressed, and nothing when
		**	released.
		*/
		if (kbd.prev_scancode != 0xe0) {

			if (kbd.prev_scancode == 0xe1 && scancode == 0x1d) {

				kbd.prev_scancode = 0x100;
				goto end_kbd_intr;
	
			} else {

				if (kbd.prev_scancode == 0x100 &&
					scancode == 0x45) {
	
					keycode = E1_PAUSE;
					kbd.prev_scancode = 0;
	
				} else {

					#ifdef KBD_REPORT_UNKNOWN

						DEBUGMSG("unknown E1 sequence");
					#endif

					kbd.prev_scancode = 0;
					goto end_kbd_intr;
				}
			}

		} else {	/* if (kbd.prev_scancode != 0xe0) ...	*/

			kbd.prev_scancode = 0;

			/*	The keyboard maintains its own internal capsi
			**	lock and num lock statuses. In caps lock mode
			**	E0 AA precedes make code and E0 2A follows
			**	break code. In num lock mode, E0 2A precedes
			**	make code and E0 AA follows break code. We do
			**	our own book-keeping, so we just ignore these.
			**
			**	For my keyboard there is no caps lock mode,
			**	but there are both Shift-L and Shift-R modes.
			**	The former mode generates E0 2A / E0 AA pairs,
			**	the latter E0 B6 / E0 36 pairs. So, we should
			**	also ignore the latter. -- aeb@cwi.nl
			*/

			if (scancode == 0x2a || scancode == 0x36) {

				goto end_kbd_intr;
			}

			if (e0_keys[scancode]) {

				keycode = e0_keys[scancode];

			} else {

				#ifdef KBD_REPORT_UNKNOWN

					DEBUGMSG("unknown scancode e0 %02x",\
						scancode);
				#endif

				goto end_kbd_intr;
			}
		}

	} else {  /* if (kbd.prev_scancode) ... */

		if (scancode >= SC_LIM) {

		        /*	This happens with the FOCUS 9000 keyboard.
			**	Its keys PF1..PF12 are reported to generate
			**	55 73 77 78 79 7a 7b 7c 74 7e 6d 6f. Moreover,
			**	unless repeated, they do not generate key-down
			**	events, so we have to zero <release> below.
			**
			**	Also, Japanese 86/106 keyboards are reported
			**	to generate 0x73 and 0x7d for \ - and \ |
			**	respectively.  
			**
			**	Also, some Brazilian keyboard is reported to
			**	produce 0x73 and 0x7e for \ ? and KP-dot,
			**	respectively.
			*/

			keycode = high_keys[scancode - SC_LIM];

			if (!keycode) {

				#ifdef KBD_REPORT_UNKNOWN

					DEBUGMSG("unknown scancode %2x", \
						scancode);
				#endif

				goto end_kbd_intr;
			}

		} else {

			 keycode = scancode;
		}
	}

	event.key.size = sizeof(ggi_key_event);
	event.key.code = keycode;

	if (release) {

		event.key.type = evKeyRelease;
		if (!test_and_clear_bit(keycode, kbd.down_keys)) {

			/*	unexpected, but this can happen: maybe this
			**	was a key release for a FOCUS 9000 PF key;
			*/
			if (keycode >= SC_LIM || keycode == 85) {

				event.key.type = evKeyPress;
				if (kbd.input.report & emKeyPress) {

					kgi_handle_input(&event);
				}
				event.key.type = evKeyRelease;
			}
		}

	} else {

		event.key.type = test_and_set_bit(keycode, kbd.down_keys)
			? evKeyRepeat : evKeyPress;
	}

	if (kbd.input.report & (1 << event.key.type)) {

		kgi_handle_input(&event);
	}

end_kbd_intr:

	/*	enable keyboard
	*/
	kbd_send_cmd(0xAE);
}

static void kbd_set_leds(ggi_uint leds)
{
	IS_CRITICAL;

	ENTER_CRITICAL;

		kbd.leds = leds;
		if (!kbd_send_data(0xed) || !kbd_send_data(leds & 0x7)) {

			/* re-enable if any errors */
			kbd_send_data(KB_ENABLE);
		}

	LEAVE_CRITICAL;
}

static ggi_uint kbd_get_leds(void)
{
	return kbd.leds;
}

#if 0	/* !!! implement */
/*	Generates sound of frequency <pitch> for <duration> msec and
**	loudness <volume>.
**
**	If pitch is 0, will turn off sound, else will turn it on for that time.
**	If duration is 0, will return immediately, else sleep for msec time,
**	then turn sound off.
**
**	We use the BEEP_TIMER vector since we're using the same method to
**	generate sound, and we'll overwrite any beep in progress. That may
**	be something to fix later, if we like.
**
**	We also return immediately, which is what was implied within the X
**	comments - KDMKTONE doesn't put the process to sleep.
*/

static void __nosound(ulong ignored)
{
	outb(inb_p(0x61) & 0xFC, 0x61);		/* disable counter 2 */
	return;
}

static void kbd_nosound(struct kgi_keyboard *k)
{
	__nosound(0);
}

static void kbd_mksound(struct kgi_keyboard *kbd, ggi_uint pitch, ggi_uint duration)
{
	static struct timer_list sound_timer = {NULL, NULL, 0, 0, __nosound};
	IS_CRITICAL;

	if (pitch > 20 && pitch < 22000) {

		pitch = 1193180 / pitch;

	} else {

		pitch = 0;	/* more than 22 kHz */
	}
 
	if (duration) {
		
		duration = (HZ*duration + 500)/1000;
		if (!duration) {

			duration = 1;
		}
	}
	
	ENTER_CRITICAL; 

		del_timer(&sound_timer);

		if (pitch) {

			io_out8(io_in8(0x61) | 3, 0x61); /* enable counter 2 */

			io_out8(0xB6, 0x43); /* for counter 2, 2 byte write */

			io_out8(pitch & 0xff, 0x42);	/* select desired HZ */
			io_out8((pitch >> 8) & 0xff, 0x42);
		
			if (duration) {

				sound_timer.expires = jiffies + duration;
				add_timer(&sound_timer);
			}

		} else {

			__nosound(0);
		}

	LEAVE_CRITICAL;
	return;
}
#endif

/*
**	initialization
*/

#ifdef KBD_DO_INIT

static int kbd_wait_for_input(void)
{
	int n;
	int status, data;

	n = TIMEOUT_CONST;
	do {
		status = inb(KB_STATUS_REG);

		/*	Wait for input data to become available.  This bit will
		**	then be cleared by the following read of the DATA reg.
		*/

		if (!(status & KB_OBF)) {
			continue;
		}

		data = inb(KB_DATA_REG);

		/*	Check to see if a timeout error has occured.  This means
		**	that transmission was started but did not complete in
		**	the normal time cycle. PERR is set when a parity error
		**	occured	in the last transmission.
		*/

		if (status & (KB_GTO | KB_PERR)) {
			continue;
		}

		return (data & 0xff);

	} while (--n);

	/*	Timed out if fell through here.
	*/
	return -1;
}

static void kbd_write(int address, int data)
{
	int status;

	do {
		/*	wait for input buffer empty
		*/
		status = inb(KB_STATUS_REG);

	} while (status & KB_IBF);

	/* write out the data */
	outb(data, address);
}

static int kbd_initialize(void)
{
	ulong flags;
	IS_CRITICAL;

	ENTER_CRITICAL;

		/*	Flush any pending input.
		*/
		while (kbd_wait_for_input() != -1) {
			continue;
		}

		/*	Test the keyboard interface. This seems to be the only
		**	way to get it going. If the test is successful a 0x55
		**	is placed in the input buffer.
		*/
		kbd_write(KB_CNTL_REG, KB_SELF_TEST);

		if (kbd_wait_for_input() != 0x55) {

			printk(KERN_ERR "Keyboard failed first self test!\n");
			restore_flags(flags);
			return 0;
		}

		/*	Perform a keyboard interface test. This causes the
		**	controller to test the keyboard clock and data lines.
		**	The results of the test are placed in the input buffer.
		*/
		kbd_write(KB_CNTL_REG, KB_SELF_TEST2);

		if (kbd_wait_for_input() != 0x00) {

			printk(KERN_ERR "Keyboard failed second self test!\n");
			restore_flags(flags);
			return 0;
		}

		/*	Enable the keyboard by allowing the keyboard clock
		**	to run.
		*/
		kbd_write(KB_CNTL_REG, KB_CNTL_ENABLE);

		/*	Reset keyboard. If the read times out then the
		**	assumption is that no keyboard is plugged into the
		**	machine. This defaults the keyboard to scan-code set 2.
		*/
		kbd_write(KB_DATA_REG, KB_RESET);

		if (kbd_wait_for_input() != KB_ACK) {

			printk(KERN_ERR "Keyboard reset failed, no ACK!\n");
			restore_flags(flags);
			return 0;
		}

		if (kbd_wait_for_input() != KB_POR) {

			printk(KERN_ERR "Keyboard reset kbd failed, no POR!\n");
			restore_flags(flags);
			return 0;
		}

		/*	now do a DEFAULTS_DISABLE always
		*/
		kbd_write(KB_DATA_REG, KB_DISABLE);

		if (kbd_wait_for_input() != KB_ACK) {

			printk(KERN_ERR "Keyboard disable failed, no ACK!\n");
			restore_flags(flags);
			return 0;
		}

		/*	Enable keyboard interrupt, operate in "sys" mode,
		**	enable keyboard (by clearing the disable keyboard bit),
		**	disable mouse, do conversion of keycodes.
		*/
		kbd_write(KB_CNTL_REG, KB_WRITE_MODE);
		kbd_write(KB_DATA_REG, KB_EKI|KBD_SYS | KB_DMS | KB_KCC);

		/*	now ENABLE the keyboard to set it scanning...
		*/
		kbd_write(KB_DATA_REG, KB_ENABLE);

		if (kbd_wait_for_input() != KB_ACK) {

			printk(KERN_ERR "Keyboard enable failed!\n");
			restore_flags(flags);
			return 0;
		}

	LEAVE_CRITICAL;
	return 1;
}
#endif /* #ifdef KBD_DO_INIT */


struct io_region kbd_io = { IO_NULL, 0x60, 16, 0xFFFF, "keyboard" };

#ifdef MODULE

ggi_uint focus = 0;
int init_module(void)
#else
#define	focus	0
int kbd_init(void)
#endif
{
	if (io_check_region(&kbd_io)) {

		printk("%s: keyboard I/O resources already in use!\n",__FILE__);
		return -EIO;
	}

#if 0
	if (check_irq(KEYBOARD_IRQ)) {

		printk("%s: failed to allocate irq.\n",__FILE__);
		return -EIO;
	}
#endif

	#ifdef KBD_DO_INIT

		if (kbd_initialize()) {

			return -EIO;
		} 
	#endif

	kgi_register_input(focus, &kbd.input);

	io_claim_region(&kbd_io);
	request_irq(KEYBOARD_IRQ, kbd_interrupt, 0, kbd.input.model, NULL);

	printk(KERN_INFO "%s %s driver\n", kbd.input.manufact, kbd.input.model);
	return EOK;
}

#ifdef MODULE

void cleanup_module(void)
{
	kgi_unregister_keyboard(&kbd);

	release_region(0x60, 16);
	free_irq(KEYBOARD_IRQ, kbd.input.model);

	#error need to do more here!
}

#endif
