/* -----------------------------------------------------------------------------
**	KGI display manager
** -----------------------------------------------------------------------------
**
**	Copyright (C) 1995-1997 by Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Steffen_Seeger
#define	VERSION		"0.00 ALPHA"

#define	DEBUG_LEVEL	0

#include <linux/version.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/malloc.h>

extern void ctrl_alt_del(void);

#include <ggi/kgi.h>

#include "default/clut.inc"
#include "default/pointer.inc"

struct kgi_focus	*kgifocus[MAX_NR_FOCUSES];
struct kgi_device	*kgidev[MAX_NR_DEVICES];
struct kgi_display	*kgidpy[MAX_NR_DISPLAYS];

static uchar display_map [MAX_NR_DEVICES];
static uchar focus_map[MAX_NR_DEVICES];

#ifdef DEBUG

static void kgi_printk_mapping_state(char *file, int line, char *action)
{
	int i;

	printk(KERN_ERR "%s:%i: %s (dev-dpy): ", file, line, action);

	for (i = 0; i < MAX_NR_DEVICES; i++) {

		if (kgidev[i]) {

			ASSERT((kgidev[i]->id == i) && kgidev[i]->dpy);
			ASSERT(kgidev[i]->dpy->id == display_map[i]);
			ASSERT(VALID_DPY_ID(display_map[i]));

			if (kgidev[i]->state & KGI_MAPPED_ON_DISPLAY) {

				printk(" %i-%i  ", i, display_map[i]);

			} else {

				printk("(%i-%i) ", i, display_map[i]);
			}
		}
	}
	printk("\n");
}

#define	PRINTK_MAP_STATE(x)	kgi_printk_mapping_state(__FILE__,__LINE__,x)

#else

#define	PRINTK_MAP_STATE(x)	do { } while (0)

#endif

ggi_clut *kgi_default_clut(ggi_mode *mode)
{
	switch(mode->graphtype) {

		case GT_TEXT16:	return &default_clut_16;

		case GT_15BIT:
		case GT_16BIT:
		case GT_24BIT:
		case GT_32BIT:	return &default_gamma;

		default:	return &default_clut_256;
	}
}

/*
**	input device administration
*/

/*	Register an input device to a focus. We keep GGI_IC_2DPOINTER devices
**	first in the linked list to tell easily if a has a pointer.
**	return values:
**		EOK	successful
**		-EINVAL	parameters invalid
*/
int kgi_register_input(int focus, struct kgi_input *dev)
{
	struct kgi_input *prev;

	if (!(dev && VALID_FOCUS_ID(focus))) {

		INVALID_PARAMS;
		return -EINVAL;
	}

	if (! kgifocus[focus]) {

		int dev;

		DEBUG2("allocating focus %i...", focus);
		if (! (kgifocus[focus] = kgi_malloc_small(sizeof(struct
			kgi_focus), GFP_KERNEL))) { 

			DEBUG2("focus %i allocation failed, no memory.", focus);
			return -ENOMEM;
		}

		memcpy(kgifocus[focus], &default_focus,
			sizeof(struct kgi_focus));
		kgifocus[focus]->id = focus;

		dev = 0;
		while ((dev < MAX_NR_DEVICES) && (focus_map[dev] != focus)) {

			dev++;
		}
		if (VALID_DEV_ID(dev)) {

			kgifocus[focus]->focus = dev;
			kgidev[dev]->state |= KGI_DEVICE_IS_FOCUSED;

		} else {

			kgifocus[focus]->focus = -1;
		}

		DEBUG2("focus %i allocated, focused on device %i.", \
			focus, kgifocus[focus]->focus);
	}

	if ((prev = kgifocus[focus]->inputs)) {

		while (prev->next && (prev->next->events & emPointer)) {

			prev = prev->next;
		}

		dev->next = prev->next;
		prev->next = dev;
		dev->focus = kgifocus[focus]->id;
		#warning implement device number handling!
		dev->id = 0;	/* !!! implement! */

	} else {

		dev->next = NULL;
		kgifocus[focus]->inputs = dev;
		dev->focus = kgifocus[focus]->id;
		#warning implement device number handling!
		dev->id = 0;	/* !!! implement */
	}

	kgifocus[focus]->flags |= (dev->events & emPointer) ? HAS_POINTER : 0;
	return EOK;
}

/*	Unregister a previously registered input device from its keyboard.
**	If dev is not valid or dev was not registered, we do nothing.
*/
void kgi_unregister_input(int focus, struct kgi_input *dev)
{
	struct kgi_input *prev;

	if (!(dev && VALID_FOCUS_ID(focus) && kgifocus[focus])) {

		INVALID_PARAMS;
		return;
	}

	if (dev != kgifocus[focus]->inputs) {

		prev = kgifocus[focus]->inputs;

		while (prev && (prev->next != dev)) {

			prev = prev->next;
		}

		if (prev) {

			prev->next = dev->next;
			dev->next  = NULL;
			dev->focus = -1;
			dev->id   = -1;

			/* If there is no 'next' device, catch that...
			*/
			if (kgifocus[focus]->inputs) {

				kgifocus[focus]->flags &= ~HAS_POINTER;
				kgifocus[focus]->flags |= (emPointer & 
					kgifocus[focus]->inputs->events) 
					? HAS_POINTER : 0;	
			} else {

				kgifocus[focus]->flags &= ~HAS_POINTER;
			}
			return;

		} else {

			/*	wrong keyboard or lost?! better report.  */
			INVALID_PARAMS;
			return;
		}

	} else {	/* if (dev != kgifocus[focus]->inputs) ... */

		kgifocus[focus]->inputs = dev->next;
		dev->next = NULL;
		dev->focus = -1;
		dev->id = -1;

		ASSERT(kgifocus[focus]->id == focus);

		if (! kgifocus[focus]->inputs) {

			DEBUG2("freeing focus %i...", focus);

			/* !!! free other (alloced) memory too (keymaps...) */
			kgi_free_small(kgifocus[focus]->inputs);
			kgifocus[focus] = NULL;
		}
	}
}

/*
**	input handling
*/

static inline void do_special(struct kgi_focus *f, ggi_event *event)
{
	if (event->any.type != evKeyPress) {

		return;
	}

	DEBUG2("doing special key %.4x", event->key.sym);

	switch(U(event->key.sym)) {

	case K_ENTER:
		return;

	case K_SH_REGS:
/*	!!!	show_regs(pt_regs);
*/		return;

	case K_SH_MEM:
		show_mem();
		event->any.type = evNothing;
		return;

	case K_SH_STAT:
		show_state();
		event->any.type = evNothing;
		return;

	case K_BREAK:
		return;

	case K_CONS:
		f->want_focus = f->last_focus;
		return;

	case K_CAPS:
	case K_NUM:
	case K_HOLD:
	case K_SCROLLFORW:
	case K_SCROLLBACK:
		return;

	case K_BOOT:
		ctrl_alt_del();
		event->any.type = evNothing;
		return;

	case K_CAPSON:
		return;

	case K_COMPOSE:
		return;

	case K_SAK:
		if (VALID_DEV_ID(f->focus) && kgidev[f->focus]) {

			do_SAK(kgidev[f->focus]->tty);
		}
		return;

	case K_DECRCONSOLE:
		{
			register int start, search;

			if (f->focus < 0) {

				start = MAX_NR_DEVICES - 1;
				search = start;

			} else {

				ASSERT(f->focus < MAX_NR_DEVICES);

				search = ((start = f->focus) > 0)
					? f->focus - 1 : MAX_NR_DEVICES - 1;
			}

			while ((search != start) && !kgidev[search] &&
				(focus_map[search] != f->id)) {

				if (--search <  0) {

					search = MAX_NR_DEVICES - 1;
				}
			}

			if (search != start) {

				f->want_focus = search;
			}
		}
		return;

	case K_INCRCONSOLE:
		{
			register int start, search;

			if (f->focus < 0) {

				start = search = 0;

			} else {

				ASSERT(f->focus < MAX_NR_DEVICES);

				search = ((start = f->focus) < MAX_NR_DEVICES-1)
					? f->focus + 1 : 0;
			}

			while ((search != start) && !kgidev[search] &&
				(focus_map[search] != f->id)) {

				if (++search >= MAX_NR_DEVICES) {

					search = 0;
				}
			}

			if (search != start) {

				f->want_focus = search;
			}
		}
		return;
			
	case K_SPAWNCONSOLE:
	case K_BARENUMLOCK:
		return;

	case K_TEXTSCREEN:
		if (VALID_DEV_ID(f->focus)
			&& (f->focus >= MAX_NR_CONSOLES)) {

			f->want_focus = f->focus - MAX_NR_CONSOLES;
			mark_bh(KEYBOARD_BH);
		}
		return;

	case K_GRAPHSCREEN:
		if (VALID_DEV_ID(f->focus)
			&& (f->focus < MAX_NR_CONSOLES)) {

			f->want_focus = f->focus + MAX_NR_CONSOLES;
			mark_bh(KEYBOARD_BH);
		}
		return;

	case K_TOGGLESCREEN:
		if (VALID_DEV_ID(f->focus)) {

			f->want_focus = (f->focus < MAX_NR_CONSOLES) 
					? f->focus + MAX_NR_CONSOLES
					: f->focus - MAX_NR_CONSOLES;
			mark_bh(KEYBOARD_BH);
		}
		return;

	default:
		DEBUG0("unknown special key %.4x", event->key.sym);
	}
}

static inline void do_modifier(struct kgi_focus *f, ggi_event *event)
{
	register ksym = U(event->key.sym);

	DEBUG2("old modifiers: %.2x %.2x %.2x %.2x, modifier key %.4x", \
		f->effect, f->normal, f->locked, f->sticky, ksym);

	/*	The effect of STICKY and LOCKED modifiers is toggled whenever
	**	they are pressed.
	*/
	if ((K_FIRST_STICKY <= ksym) && (ksym <= K_LAST_STICKY)) {

		NOTICE2("sticky modifer key.");

		if (event->key.type == evKeyPress) {

			f->sticky ^= (1 << KVAL(ksym - K_FIRST_STICKY));	
		}
	}

	if ((K_FIRST_LOCKED <= ksym) && (ksym <= K_LAST_LOCKED)) {

		NOTICE2("locked modifier key.");

		if (event->key.type == evKeyPress) {

			f->locked ^= (1 << KVAL(ksym - K_FIRST_LOCKED));
		}
	}

	/*	If a NORMAL modifier is pressed at least once, it goes into
	**	effect and clears the effect of the corresponding LOCKED
	**	modifier. Thus there is no effect in the EFFECTIVE modifers.
	**	The effect of the NORMAL modifier is not cleared until the
	**	last is released (if pressed multiple times), which cleares
	**	the effect of the corresponding LOCKED modifier too.
	*/
	if ((K_FIRST_NORMAL <= ksym) && (ksym <= K_LAST_NORMAL)) {

		register ggi_uint mask = 1 << KVAL(ksym - K_FIRST_NORMAL);

		NOTICE2("normal modifier key.");

		if (event->key.type == evKeyPress) {

			f->down_mod[KVAL(ksym - K_FIRST_NORMAL)]++;
			f->normal |=  mask;
			f->locked &= ~mask;

		} else {

			ASSERT(event->key.type == evKeyRelease);
			ASSERT(f->down_mod[KVAL(ksym - K_FIRST_NORMAL)] > 0);

			if (! --(f->down_mod[KVAL(ksym - K_FIRST_NORMAL)])) {

				f->normal &= ~mask;
				f->locked &= ~mask;
			}
		}
	}

	f->effect = f->normal ^ f->locked ^ f->sticky;

	DEBUG2("new modifers: %.2x %.2x %.2x %.2x", f->effect, f->normal,\
		f->locked, f->sticky);
}

static inline void do_ascii(struct kgi_focus *f, ggi_event *event)
{
	int base, value;

	DEBUG2("doign ascii key %.4x", event->key.sym);

	if (event->key.type == evKeyRelease) {

		return;
	}

	if (KVAL(event->key.sym) < 10) {

		value = KVAL(event->key.sym);
		base = 10;

	} else {

		value = KVAL(event->key.sym) - 10;
		base = 16;
	}

	f->npadch = (f->npadch == -1) ? value : (f->npadch*base + value);

	if (base == 16) {

		f->npadch &= 0xFFFF;

	} else {

		if (f->npadch > 100000) {

			f->npadch %= 100000;
		}
	} 
}

static inline void do_action(struct kgi_focus *f, ggi_event *event)
{
	ggi_uint val = KVAL(event->key.sym);
	ggi_uint typ = KTYP(event->key.sym);

	if ((1 << event->any.type) & ~(emKeyPress | emKeyRelease)) {

		return;
	}

	DEBUG2("key %s, code %.2x, sym %.2x", event->key.type == evKeyPress\
		? "down" : "up", event->key.code, event->key.sym);

	if (typ >= 0xF0) {

		typ -= 0xF0;

		switch (typ) {

		case KT_SPEC:
			if (val < K_LAST_SPEC) {

				do_special(f, event); 
			}
			return;

		case KT_CONS:
			if ((val < MAX_NR_DEVICES) &&
				(event->key.type == evKeyPress)) {

				f->want_focus = val;
				mark_bh(KEYBOARD_BH);
			}
			return;

		case KT_SHIFT:
			if (val < K_LAST_MODIFIER) {

				do_modifier(f, event);
			}
			return;
	
		case KT_ASCII:
			if (val < K_LAST_ASCII) {

				do_ascii(f, event);
			}
			return;
		}
	}
}

void kgi_handle_input(ggi_event *event)
{
	struct kgi_focus *f;
	ggi_uint mask = 1 << event->any.type;

	ASSERT(event->any.focus < MAX_NR_FOCUSES);

	/* Protect against bogus events. */
	if (event->any.focus >= MAX_NR_FOCUSES) return;

	f=kgifocus[event->any.focus];	
	ASSERT(f!=NULL);

	if (mask & (emKey | emPointer)) {

		event->key.effect = f->effect;
		event->key.normal = f->normal;
		event->key.locked = f->locked;
		event->key.sticky = f->sticky;
	}

	if ((mask & emKey) && (event->key.code < 0x8000)) {

		struct kgi_keymap *k = f->kmap;
		unicode *map = (f->effect < k->keymap_size) 
				? k->keymap[f->effect] : NULL;

		if (map) {

			event->key.sym = ((k->keymin <= event->key.code) &&
				(event->key.code <= k->keymax))
				? map[event->key.code - k->keymin] : K_VOID;

		} else {

			map = k->keymap[0];

			if ((k->keymin <= event->key.code) &&
				(event->key.code <= k->keymax) &&
				(KTYP(U(map[event->key.code - k->keymin]))
				== KT_SHIFT)) {

				NOTICE2("using plainmap to catch shift");
				event->key.sym = map[event->key.code-k->keymin];

			} else {

				NOTICE2("no shift translation");
				event->key.sym = K_VOID;
			}
		}

		do_action(f, event);
	}

	if (VALID_DEV_ID(f->focus)) {

		register struct kgi_device *dev = kgidev[f->focus];

		if (dev && (mask & dev->event_mask)) {

			(dev->handle_event)(dev, event);
		}
	}
}

/*	This routine runs with all interrupts enabled and does all the things
**	that may have to be done in response to an handled event, but may take
**	a reasonable long time.
*/
static void kgi_bottomhalf(void)
{
	int i;

	for (i = 0; i < MAX_NR_FOCUSES; i++) {

		if (kgifocus[i]) {

			ASSERT(VALID_DEV_ID(kgifocus[i]->want_focus) ||
				(kgifocus[i]->want_focus == -1));

			if (kgifocus[i]->want_focus >= 0) {

				kgifocus[i]->last_focus = kgifocus[i]->focus;
				kgi_focus_device(kgifocus[i]->want_focus);
			}
			kgifocus[i]->want_focus = -1;
		}
	}
}

char *kgi_function_key_string(int focus, ggi_uint fnkey)
{
	ASSERT(fnkey < 256);
	if (VALID_FOCUS_ID(focus) && kgifocus[focus]) {

		return (fnkey < kgifocus[focus]->kmap->fn_str_size)
			? kgifocus[focus]->kmap->fn_str[fnkey] : NULL;
	} else {

		return NULL;
	}
}

unicode kgi_combine_dead(int focus, ggi_uint dead, ggi_uint key)
{
	#warning implement dead keys!
	DEBUG1("combining dead %.4x with %.4x", dead, key);
	return K_VOID;
}

/*
**	display <-> device mapping control
*/

/*	Map the device <dev> on it's display and focus the associated
**	keyboard on <dev> (if there is one). The device <dev> must be
**	attached to a display, meaning there is a valid reference in kgidev[]
*/
void kgi_focus_device(ggi_uint id)
{
	IS_CRITICAL;
	struct kgi_device *dev;
	struct kgi_focus *f;

	if (! (VALID_DEV_ID(id) && (dev = kgidev[id]))) {

		INVALID_PARAMS;
		return;
	}

	ASSERT(dev->id == id);
	ASSERT(display_map[id] == dev->dpy->id);

	ENTER_CRITICAL;

		if ((f = kgifocus[focus_map[id]])) {

			if (kgidev[f->focus]) {

				kgi_hide_gadgets(kgidev[f->focus]);
				kgidev[f->focus]->state &=
					~KGI_DEVICE_IS_FOCUSED;
			}

			f->focus = dev->id;
			f->x	= dev->pointer.x;
			f->y	= dev->pointer.y;
			f->minx	= dev->ptr_min.x;
			f->miny = dev->ptr_min.y;
			f->maxx = dev->ptr_max.x;
			f->maxy = dev->ptr_max.y;

			dev->state |= KGI_DEVICE_IS_FOCUSED;
		}

		if (dev->dpy->dev != dev) {

			kgi_map_device(dev);

		} else {

			kgi_show_gadgets(dev);
		}

	LEAVE_CRITICAL;
}

/*
 *	display handling
 */

/*\
|*|	NOTE: the kgi_????_gadgets() functions must not be called when
|*|	dev is not mapped on dev->dpy!!
\*/

/*	Hide the cursor and pointer, even if they are in hardware.
*/
void kgi_hide_gadgets(struct kgi_device *dev)
{
	IS_CRITICAL;
	struct kgi_display *dpy = dev->dpy;

	/* Commented out because of annoyance */
/*	ASSERT(dev->state & KGI_MAPPED_ON_DISPLAY); */

	ENTER_CRITICAL;

		if ((dev->state & KGI_POINTER_SHOWN) &&
			(dpy->hide_pointer)) {

			(dpy->hide_pointer)(dpy);
			dev->state &= ~(KGI_POINTER_FB_MODIFIED
					| KGI_POINTER_SHOWN);
		}

		if ((dev->state & KGI_CURSOR_SHOWN) &&
			(dpy->hide_cursor)) {

			(dpy->hide_cursor)(dpy);
			dev->state &= ~(KGI_CURSOR_FB_MODIFIED
					| KGI_CURSOR_SHOWN);
		}

	LEAVE_CRITICAL;
}

/*	Undo fb modifications, does nothing if we have hardware cursor/pointer.
*/
void kgi_undo_gadgets(struct kgi_device *dev)
{
	IS_CRITICAL;
	struct kgi_display *dpy = dev->dpy;

/*
**	This can cause an infinite loop since it is
**	called by console_print - disabled (emarty)
**
**	ASSERT(dev->state & KGI_MAPPED_ON_DISPLAY);
*/

	ENTER_CRITICAL;

		if ((dev->state & KGI_POINTER_FB_MODIFIED) &&
			(dpy->undo_pointer)) {

			(dpy->undo_pointer)(dpy);
			dev->state &= ~(KGI_POINTER_FB_MODIFIED
					| KGI_POINTER_SHOWN);
		}

		if ((dev->state & KGI_CURSOR_FB_MODIFIED) &&
			(dpy->undo_cursor)) {

			(dpy->undo_cursor)(dpy);
			dev->state &= ~(KGI_CURSOR_FB_MODIFIED
					| KGI_CURSOR_SHOWN);
		}

	LEAVE_CRITICAL;
}

/*	Update cursor and pointer, undo fb modifications before.
*/
void kgi_show_gadgets(struct kgi_device *dev)
{
	IS_CRITICAL;
	struct kgi_display *dpy = dev->dpy;

/*
**	This can cause an infinite loop since it is
**	called by console_print - disabled (emarty)
**
**	ASSERT(dev->state & KGI_MAPPED_ON_DISPLAY);
*/

	ENTER_CRITICAL;

		if ((dev->state & KGI_POINTER_FB_MODIFIED) &&
			(dpy->undo_pointer)) {

			(dpy->undo_pointer)(dpy);
		}

		if ((dev->state & KGI_CURSOR_FB_MODIFIED) &&
			(dpy->undo_cursor)) {

			(dpy->undo_cursor)(dpy);
		}

		if ((dev->state & KGI_CURSOR_TO_SHOW) &&
			(dev->state & KGI_DEVICE_IS_FOCUSED) &&
			(dpy->show_cursor)) {

			(dpy->show_cursor)(dpy, dev->cursor.x, dev->cursor.y);
			dev->state |= KGI_CURSOR_SHOWN | (dpy->undo_cursor
				? KGI_CURSOR_FB_MODIFIED : 0);
		}

		if ((dev->state & KGI_POINTER_TO_SHOW) &&
			(dev->state & KGI_DEVICE_IS_FOCUSED) &&
			(dpy->show_pointer)) {

			(dpy->show_pointer)(dpy, dev->pointer.x,dev->pointer.y);
			dev->state |= KGI_POINTER_SHOWN | (dpy->undo_pointer
				? KGI_POINTER_FB_MODIFIED : 0);
		}

	LEAVE_CRITICAL;
}


/*	Check if a mode is valid. This returns a valid <mode> and EOK if
**	the mode can be done. NOTE: <cmd> must be either CMD_PROPOSE or
**	CMD_CHECK. <dpy> and <mode> must be valid.
*/
inline int kgi_check_mode(struct kgi_display *dpy, struct kgi_mode *mode, 
			enum kgi_tm_cmd cmd)
{
	ASSERT(dpy);
	ASSERT(mode);
	ASSERT((cmd == CMD_PROPOSE) || (cmd == CMD_CHECK));

	if (mode->request.frames >= NR_FB) {

		return -ENOMEM;
	}

	return (dpy->check_mode)(dpy, mode, cmd);
}

/* !!!	Check, adopt calling interface to accept a device instead of a display.
*/
inline void kgi_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	ASSERT(dpy);
	ASSERT(mode);

	if (memcmp(&(dpy->mode), mode, sizeof(struct kgi_mode))) {

		(dpy->set_mode)(dpy, mode);
	}
}

/*	Set the pointer window to the given minimal and maximal values.
*/
void kgi_set_pointer_window(struct kgi_device *dev, ggi_uint x1, ggi_uint x2,
	ggi_uint y1, ggi_uint y2)
{
	struct kgi_focus *f;

	ASSERT(dev);
	ASSERT(VALID_DEV_ID(dev->id));
	ASSERT(focus_map[dev->id] < MAX_NR_FOCUSES);

	f = kgifocus[focus_map[dev->id]];

	ASSERT(x1 <= x2);
	ASSERT(y1 <= y2);
	ASSERT(x2 <= dev->mode.visible.x);
	ASSERT(y2 <= dev->mode.visible.y);
	ASSERT(!(dev->state & KGI_POINTER_TO_SHOW));

	dev->ptr_min.x = x1;
	dev->ptr_min.y = y1;
	dev->ptr_max.x = x2;
	dev->ptr_max.y = y2;

	if (dev->pointer.x < x1)	dev->pointer.x = x1;
	if (x2 <= dev->pointer.x)	dev->pointer.x = x2 - 1;
	if (dev->pointer.y < y1)	dev->pointer.y = y1;
	if (y2 <= dev->pointer.y)	dev->pointer.y = y2 - 1;

	if (f && (f->focus == dev->id)) {

		f->x = dev->pointer.x;
		f->y = dev->pointer.y;
		f->minx = dev->ptr_min.x;
		f->miny = dev->ptr_min.y;
		f->maxx = dev->ptr_max.x;
		f->maxy = dev->ptr_max.y;
	}
}

/*	Set the pointer coordinates to the given 
*/
void kgi_set_pointer(struct kgi_device *dev, ggi_uint x, ggi_uint y)
{
	dev->pointer.x = (x < dev->ptr_min.x) ? dev->ptr_min.x : 
		((x < dev->ptr_max.x) ? x : (dev->ptr_max.x - 1));

	dev->pointer.y = (y < dev->ptr_min.y) ? dev->ptr_min.y :
		((y < dev->ptr_max.y) ? y : (dev->ptr_max.y - 1));
}

/*	Set the visible area of the current frame to originate from the pixel
**	(x,y). <dev> needs to be attached to a display. If the display supports
**	set_origin then dev->org[frame] is updated. If the device is mapped, the
**	hardware is setup too.
*/
void kgi_set_origin(struct kgi_device *dev, ggi_uint frame, ggi_uint x, ggi_uint y)
{
	register struct kgi_display *dpy = dev->dpy;

	ASSERT(dpy);
	ASSERT(frame < NR_FB);

	if (dpy->set_origin) {

		dev->org[frame].x = x;
		dev->org[frame].y = y;

		if (dev->state & KGI_MAPPED_ON_DISPLAY) {

			(dpy->set_origin)(dpy, x, y);
		}
	}
}

/*	Set the split line for frame 0. dev needs to be attached to a display.
**	If the display supports the split line feature, dev->split is updated.
**	If the device is mapped, the hardware is setup too.
*/
void kgi_set_split(struct kgi_device *dev, ggi_uint line)
{
	register struct kgi_display *dpy = dev->dpy;

	if (dpy->set_split) {

		dev->split = line;

		if (dev->state & KGI_MAPPED_ON_DISPLAY) {

			(dpy->set_split)(dpy, line);
		}
	}
}

/*	Register <dpy> under <id>. If <id> is negative, we search for the first
**	free one. Basically there are two cases:
**	(a) we register a full driver and might have to replace a boot driver.
**	(b) we register a boot driver and might have to replace a full driver.
**	Other calls are invalid. If a boot driver is replaced, we destroy the
**	replaced struct kgi_display using text16_free_display().
**	return values:
**		EOK	successful
**		-EINVAL	calling parameters are invalid.
**		-ENOMEM	no free <id> found
**		-EBUSY	the new driver can't do the modes served by the old one
*/
int kgi_register_display(struct kgi_display *dpy, int id)
{
	IS_CRITICAL;
	int dev;

	if (id < 0) {

		for (id = 0; (id < MAX_NR_DISPLAYS) && kgidpy[id]; id++) {
		}

		if (MAX_NR_DISPLAYS <= id) {

			return -ENOMEM;
		}
	}

	if (   !VALID_DPY_ID(id) || !dpy
	    || (kgidpy[id] &&  kgidpy[id]->card &&  dpy->card)
	    || (kgidpy[id] && !kgidpy[id]->card && !dpy->card)) {

		INVALID_PARAMS;
		return -EINVAL;
	}

	/*	If a display driver handles this hardware already, we have to
	**	make sure the new driver can support the same modes for the
	**	devices attached to it.
	*/
	if (kgidpy[id]) {

		for (dev = 0; dev < MAX_NR_DEVICES; dev++) {

			if ((display_map[dev] == id) && kgidev[dev]) {

				struct kgi_mode mode;

				ASSERT(kgidev[dev]->dpy == kgidpy[id]);

				mode.request = kgidev[dev]->mode;

				ASSERT(mode.request.virt.x && mode.request.virt.y);

				if (kgi_check_mode(dpy, &mode, CMD_PROPOSE)) {

					DEBUG2("dpy can't do mode for dev %i!",\
						dev);
					return -EBUSY;
				}

				#define	TEST(x) (mode.request.x !=\
							kgidev[dev]->mode.x)

				if (   TEST(visible.x) || TEST(visible.y)
				    || TEST(virt.x) || TEST(virt.y)
				    || TEST(text.x)    || TEST(text.y)
				    || TEST(graphtype) || TEST(frames)) {

					DEBUG2("fblayout change (dev %i)!\n",\
						dev);
					return -EBUSY;
				}

				#undef TEST
			}
		}
	}

	/* !!!	temporary, later we have to check exactly for conflicts
	*/
	dpy->conflict_setup = 0xFFFFFFFF;
	dpy->conflict_text  = 0xFFFFFFFF;
	dpy->conflict_graph = 0xFFFFFFFF;
	dpy->id = id;
	dpy->refcnt = 0;
	dpy->dev = NULL;

	/*	Now <dpy> is prepared and we know the devices can be attached.
	**	Thus we can replace the display. NOTE: THIS CODE RELIES ON
	**	kgi_detach_device() NOT ACCESSING kgidpy[dpy]!
	*/
	ENTER_CRITICAL;
	{
		struct kgi_display *old = kgidpy[id];
		struct kgi_device *mapped = old ? old->dev : NULL;
		struct kgi_device *tmp;

		if (mapped) {

			kgi_unmap_device(mapped);
		}

		kgidpy[id] = dpy;

		for (dev = 0; dev < MAX_NR_DEVICES; dev++) {

			tmp = kgidev[dev];

			if ((display_map[dev] == id) && tmp) {

				kgi_detach_device(tmp);
				if (kgi_attach_device(tmp)) {

					/*	We should never get here,
					**	because we checked that
					**	attachment would go well.
					*/
					INTERNALERROR;
				}
			}
		}

		if (mapped) {

			kgi_map_device(mapped);
		}

		DEBUG2("display %s %s registered.", dpy->card->manufact, \
			dpy->card->model);

		if (old && !old->card) {

			ASSERT(old->refcnt == 0);
			text16_free_display(old);
		}
	}
	LEAVE_CRITICAL;

	return EOK;
}


/*	Unregister a previously registered display. There might be an alias
**	display driver provided that handles devices at minimal functionality.
**	Note that this alias must be allocated using text16_malloc_display().
**	return values:
**		EOK	successful
**		-EBUSY	no alias provided but referenced or the alias provided
**			is invalid/cannot do the modes of attached devices.
**		-EINVAL	invalid parameters
*/
int kgi_unregister_display(struct kgi_display *dpy, struct kgi_display *alias)
{ 
	if (!(dpy && VALID_DPY_ID(dpy->id) && (kgidpy[dpy->id] == dpy))) {

		INVALID_PARAMS;
		return -EINVAL;
	}

	if (alias) {

		int err;

		if (alias->card) {

			INVALID_PARAMS;
			return -EINVAL;
		}

		err = kgi_register_display(alias, dpy->id);

		ASSERT(err != -EINVAL);
		ASSERT(dpy->refcnt == 0);

		return err;

	} else {
	
		if (dpy->refcnt) {

			return -EBUSY;
		}

		ASSERT(!dpy->dev);

		kgidpy[dpy->id] = NULL;
		(dpy->reset)(dpy);
	}

	dpy->id     = -1;
	dpy->dev    = NULL;
	dpy->refcnt = 0;

	return EOK;
}

#if NOT_IMPLEMENTED_YET
void kgi_enable_display(struct kgi_display *dpy, int setup)
{
  /*\
  |*|	enable dpy for access (setup == 0) or setup (setup != 0).
  |*|	if dpy is NULL, disable all registered displays (detection 
  |*|	of unregistered)
  \*/

  static ggi_uint enabled = 0;
  ggi_uint conflict, check = 0, mask = 1;

  if (dpy)
    {
      if ((1 << dpy->id) & enabled)
         return;			/* already enabled		*/

       conflict = setup ? dpy->conflict_setup 
 		   : ((dpy->mode.request.graphtype == GT_TEXT16)
				? dpy->conflict_text 
				: dpy->conflict_graph);
    }
   else
    conflict = 0xFFFFFFFF;

  conflict &= enabled;			/* only enabled and conflicting	*/

  while (check < 32)
    {
      if (mask & conflict)
        {
          display[check]->disable(display[check]);
          enabled ^= mask;
        }
      check++;
      mask += mask;
    }

  if (dpy)
    {
      dpy->enable(dpy);
      enabled |= (1 << dpy->id);
    }
}
#endif

static void set_fb_param(struct kgi_device *dev, kgi_fb_ops *ops)
{
	int cnt = dev->mode.frames;
	int sizex = dev->mode.virt.x;
	int sizey = dev->mode.virt.y;
	int bpp;

	if (   (dev->mode.graphtype == GT_TEXT16)
	    || (dev->mode.graphtype == GT_TEXT32)) {

		sizex /= dev->mode.text.x;
		sizey /= dev->mode.text.y;
		bpp   = (dev->mode.graphtype == GT_TEXT16)
				? MAP_16_BPP : MAP_32_BPP;
	} else {

		printk(KERN_ERR "set_fb_param(): can't handle graphic mode.\n");
		return;
	}

	memset(dev->fb, 0, sizeof(dev->fb));

	while (cnt--) {

		dev->fb[cnt].size.x	= sizex;
		dev->fb[cnt].size.y	= sizey;
		dev->fb[cnt].bpp	= bpp;
	}
}

/*
**	device handling
*/

/*	Attach a device to its display and keyboard. The keyboard may be
**	missing, but the display has to be registered. Only attached devices
**	~have a reference in kgidev[].
**	return values:
**		EOK	successful
**		-EINVAL	invalid parameters or device already attached
**		-ENODEV	no display to attach to.
*/
int kgi_attach_device(struct kgi_device *dev)
{
	struct kgi_display *dpy;
	struct kgi_mode mode;
	int i, focus;

	ASSERT(dev);

	if (!VALID_DEV_ID(dev->id) || kgidev[dev->id]) {

		INVALID_PARAMS;
		ASSERT(kgidev[dev->id] == dev);
		ASSERT(VALID_DEV_ID(dev->id));
		return -EINVAL;
	}

	dpy = kgidpy[display_map [dev->id]];
	focus = focus_map[dev->id];
	ASSERT(VALID_FOCUS_ID(focus));

	if (! dpy) {

		DEBUG2("no display to attach device %i to.", dev->id);
		return -ENODEV;
	}

	if (dev->mode.virt.x || dev->mode.virt.y) {

		mode.request = dev->mode;

	} else {

		mode.request = *(dpy->defmode);
		mode.request.virt.y = 0;
	}

	if (kgi_check_mode(dpy, &mode, CMD_PROPOSE)) {

		NOTICE1("requested mode not possible.");
		return -EINVAL;
	} 

	if (! (dev->mode.virt.x && dev->mode.virt.y)) {

		/*	<dev> is attached the first time, thus we have to
		**	set some more fields to their default values.
		*/
		dev->mode  = mode.request;
		dev->split = dev->mode.virt.y;
		dev->font  = *(kgi_default_font(&mode.request));
		dev->ptr_shape = &default_pointer;

		for (i = 0; i < dev->mode.frames; i++) {

			dev->lut[i] = *(kgi_default_clut(&dev->mode));
			dev->org[i].x = dev->org[i].y = 0;
		}

		set_fb_param(dev, mode.mem);
	}

	dev->dpy = dpy;
	dev->focus = focus;
	kgidev[dev->id] = dev;
	dpy->refcnt++;
	dev->refcnt++;

	dev->mem = mode.mem;
	for (i = 0; i < dev->mode.frames; i++) {

		dev->fb[i].ops = mode.mem;
	}

	return EOK;
}

/*	Detach a previously attached device. The device must not be mapped
**	NOTE:	THIS ROUTINE MUST NOT ACCESS (OR CALL ROUTINES THAT ACCESS)
**		kgidpy[dev->dpy->id]!! ONLY USE dev->dpy!!
*/
void kgi_detach_device(struct kgi_device *dev)
{
	int i;

	if (!dev || !VALID_DEV_ID(dev->id) || (kgidev[dev->id] != dev)) {

		INVALID_PARAMS;
		return;
	}

	ASSERT(!(dev->state & KGI_MAPPED_ON_DISPLAY));

	dev->dpy->refcnt--;
	dev->refcnt--;
	dev->dpy = NULL;
	dev->focus = -1;
	kgidev[dev->id] = NULL;

	dev->mem = NULL;
	for (i = 0; i < NR_FB; i++) {

		dev->fb[i].ops = NULL;
	}
}

/*	unmap the device dev from the display it is mapped on.
*/
void kgi_unmap_device(struct kgi_device *dev)
{
	ASSERT(dev && VALID_DEV_ID(dev->id) && (kgidev[dev->id] == dev));
	ASSERT(dev->dpy->dev == dev);
	ASSERT(dev->state & KGI_MAPPED_ON_DISPLAY);

	kgi_hide_gadgets(dev);
	(dev->save)(dev);
	dev->dpy->dev = NULL;
	dev->state &= ~KGI_MAPPED_ON_DISPLAY;

	#warning tell the application it was unmapped!!!
	PRINTK_MAP_STATE("unmap");
}

/*	map dev on the display it is attached to.
*/
void kgi_map_device(struct kgi_device *dev)		
{
	struct kgi_mode mode;
	register struct kgi_display *dpy = dev->dpy;
	struct kgi_focus *f = kgifocus[focus_map[dev->id]];

	ASSERT(dev && VALID_DEV_ID(dev->id) && (kgidev[dev->id] == dev));
	ASSERT(dev->dpy->dev != dev);
	ASSERT(! (dev->state & KGI_MAPPED_ON_DISPLAY));

	if (f && (f->focus == dev->id)) {

		dev->state |= KGI_DEVICE_IS_FOCUSED;
	}

	if (dpy->dev) {

		if (dpy->dev == dev) {

			return;

		} else {

			kgi_unmap_device(dpy->dev);
		}
	}

	dev->state |= KGI_MAPPED_ON_DISPLAY;
	dpy->dev = dev;

	mode.request = dev->mode;
	kgi_check_mode(dpy, &mode, CMD_PROPOSE);
	kgi_set_mode(dpy, &mode);

	if (dpy->set_font) {

		(dpy->set_font)(dpy, &(dev->font), 
			0, dev->font.info->positions - 1);
	}
	if (dpy->set_frame) {

		(dpy->set_frame)(dpy, dev->frame);
	}
	if (dpy->set_split) {

		(dpy->set_split)(dpy,dev->frame?dev->mode.virt.y:dev->split);
	}
	if (dpy->set_clut) {

		ggi_clut *clut = dev->lut + dev->frame;

		(dpy->set_clut)(dpy, clut, 0, clut->size); /*??? size correct?*/
	}
	if (dpy->set_origin) {

		(dpy->set_origin)(dpy,	dev->org[dev->frame].x, 
					dev->org[dev->frame].y);
	}
	if (dpy->set_pointer_shape && dev->ptr_shape) {

		(dpy->set_pointer_shape)(dpy, dev->ptr_shape);
	}

	(dev->restore)(dev);

	if (f && (f->focus == dev->id)) {

		kgi_show_gadgets(dev);
	}

	#warning tell the application it was mapped!!!
	PRINTK_MAP_STATE("map");
}

/*	kgi_alloc_device allocates and initializes the device <num>. It uses
**	the given <priority> to allocate memory. Initialization will fill all
**	fields with zeros.
**	return values:
**		NULL	out of memory
**		other	success
*/
void *kgi_alloc_device(ggi_uint num, int priority)
{
	ggi_uint type = (num < MAX_NR_CONSOLES)
				? KGI_CONSOLE_DEV
				: KGI_GRAPHIC_DEV;

	ggi_uint size = (type == KGI_CONSOLE_DEV)
				? sizeof(struct kgi_console)
				: sizeof(struct kgi_graphic);

	struct kgi_device *dev = kgi_malloc_small(size, priority);

	if (dev) {

		memset(dev, 0, size);
		dev->type = type;
		dev->id   = num;
		/* dev->refcnt is set to 0 already! */
	}

	return dev;
}

/*	Free the kgi_device <dev>.
**	NOTE:	The console or graphic device drivers have to free all
**	NOTE:	dynamically allocated memory they associate with the device
**	NOTE:	and to unmap and detach the device *before* calling
**	NOTE:	kgi_free_device()!! Not doing not so will result in hard
**	NOTE:	to find memory leaks!
*/
void kgi_free_device(struct kgi_device *dev)
{
	if (!dev || dev->refcnt) {

		INVALID_PARAMS;
		return;
	}

	kgi_free_small(dev);
}


/*
**	KGI manager initialization
*/

/*	All kernel services need to be accessible when kgi_init is called,
**	especially PCI-scanning, kgi_malloc_*() services. We first
**	initialize global variables and then get the services running.
*/
void kgi_init(void)
{
	memset(display_map, 0, sizeof(display_map));
	memset(focus_map, 0, sizeof(focus_map));

	memset(kgidpy, 0, sizeof(kgidpy));
	memset(kgidev, 0, sizeof(kgidev));
	memset(kgifocus, 0, sizeof(kgifocus));

	if (display_init() > 1) {

		display_map[MAX_NR_CONSOLES + 3] = display_map[3] =
		display_map[MAX_NR_CONSOLES + 4] = display_map[4] =
		display_map[MAX_NR_CONSOLES + 5] = display_map[5] = 1;
	}

	dev_console_init();
	dev_graphic_init();
	/* !!!	vcs_init();	*/

	init_bh(KEYBOARD_BH, kgi_bottomhalf);

	if (kbd_init() > 1) {

		focus_map[MAX_NR_CONSOLES + 3] = focus_map[3] =
		focus_map[MAX_NR_CONSOLES + 4] = focus_map[4] =
		focus_map[MAX_NR_CONSOLES + 5] = focus_map[5] = 1;
	}
}

void do_blank_screen(void)
{
}

void do_unblank_screen(void)
{
}

