/* ---------------------------------------------------------------------------
 * ggisetfont.c, a Linux console font loader by Jon Taylor 
 * (taylorj@csus.edu), based on setfont.c by Albert Calahan,
 * albert@ccs.neu.edu (also adc@coe.neu.edu & acahalan@lynx.neu.edu)
 * Based on setfont.c by Eugene Crosser & aeb.
 *
 * Version 0.5
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 * ---------------------------------------------------------------------------
 */

/* ---------------------------------------------------------------------------
 * This program should read these fonts:
 * Raw files with no header
 * .psf files with 4-byte header
 * .cpi multifont files, like those included with DOS
 * .cp files with 1 font in several sizes
 * *** Font must be 8x6 - 8x32 with 256 chars ***
 * ---------------------------------------------------------------------------
 */

/*
 * Help!
 * How do I convert "Code Page 161", "Latin 1", "France", and "ISO 03" ?
 */

#include <stdio.h>
#include <memory.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/kd.h>

#include <ggi/ggi.h>
#include <ggi/font.h>
#include <ggi/ggi_commands.h>
#include "../patches/Linux/linux/drivers/char/default/font.inc"
#include "../patches/Linux/linux/drivers/char/default/ibm-cp437.08x08.inc"

char *progname;

int error;		/* There was an error */
int dbl;		/* True if font gets doubled */
int cp;			/* Requested code page */
char *ifil;		/* Name of input file */
int height = 16;	/* Requested character height */
unsigned char buffer[10000];


extern int opterr, optind;
extern char *optarg;

int  filesize(int fd);
void *readfont(struct kgi_font *font, FILE *fp);
int  search_cp(FILE *fp);
int  seekfont(int height, FILE *fp);
int  uploadfont(struct kgi_font *font);


void main(int argc, char *argv[])
{
	struct kgi_font font = font_8x8;
	char *fontfile;
	FILE *fp;
	void *bufptr;
	
	fontfile = argv[1];
	
	printf ("Opening font file %s\n", fontfile);

	if ((fp = fopen(fontfile, "r")) == NULL) {
		printf("Cannot open font file %s.\n", fontfile);
		exit(0);
	}

	font.height = seekfont(height, fp);
	printf("Font.height = %i\n", font.height);
	bufptr = readfont(&font, fp); 
	font.data = bufptr;
	font.height = 16;
	font.dy = 16;
	font.base = 11;
/*	font = font_8x8; */
	uploadfont(&font);
}

void *readfont(struct kgi_font *font, FILE *fp){
		
	if (fread(buffer, height * 256, 1, fp) != 1) 
	{
		perror ("Cannot read font from file");
		exit (1);
	}
	return buffer;
	
}

/*********************************************/
int seekfont(int height, FILE *fp){
	int hdr;
	int size;

	size = filesize(fp->_fileno);
	
	if (size == 9780) {  /* This is a codepage file */
		/* Normal code page: first 34 bytes main header,
		 then 8x16, 8x14, and 8x8 fonts each preceded by 6 byte subheader */
		return(search_cp(fp));
	 }
	
	if((size > 9780) && (size < 65537)){  /* This is a big file, might be .cpi */
		int cpnum;
		char cpihdr[25];
		char cpimagic[] = "\377FONT   ";
		
		if (fread(cpihdr, 25, 1, fp) != 1) {
			perror("Error reading input font .cpi header");
			return(-1);
		}
		if(strncmp(cpimagic,cpihdr,8) != 0){ /* What the hell? */
			printf("Unknown font file type.\n");
			return(-1);
		}
		cpnum = cpihdr[23]; /* number of .cp files inside .cpi */
		printf("%i code pages in this .cpi file.\n",cpnum);
		/* start a loop here checking each .cp part for a cp/unit match */
		while(cpnum){
			int tmpunits;
			tmpunits=search_cp(fp);
			cpnum--;
		}
		return(0);
	}
	
	/* Not a normal codepage file, so do basic error checking & setup. */
	hdr = (size & 255);
	height = (size >> 8);
	
	if (height < 6 || height > 32) {
		/* Although font sizes less than 6 are valid, they are stupid. */
		printf("Bad input file size.\n");
		return(-1);
	}
	 
	if (hdr == 4) {                       /* .psf file */
		char psfhdr[4];
		if (fread(psfhdr, 4, 1, fp) != 1) {
			perror("Error reading header input font");
			return(-1);
		}
		/* note: this depends on endianness */
		if (psfhdr[1] != 0x04 || psfhdr[0] != 0x36) {
			printf("Unrecognized font format\n");
			return(-1);
		}
		if (psfhdr[2] != 0) {
			printf("Unsupported psf file mode\n");
			return(-1);
		}
		if(size != hdr + 256 * psfhdr[3]) {
			printf("Input file: bad length\n");
			return(-1);
		}
		return(height);
	}
	 
	if (hdr == 0) return (height);  /* Must be raw file, needs no work */
	
	if (hdr == 40) {               /* .cp file with only one font */
		return(search_cp(fp));
	}
	
	/* add more file types here */
	printf("Unknown input file type.\n");
	return(-1);
}
 
int search_cp(FILE *fp)
{
	char hdr[34];
	unsigned short numfonts;
	unsigned short fontsleft;
	unsigned short thispage;
	bool skip;
	unsigned short *sptr;
	sptr=(short *)hdr;
	if(fread(hdr, 34, 1, fp) != 1) {
		perror("Error reading .cp header");
		return(0);
	}
	numfonts = sptr[15];
	skip = 0;
	thispage = sptr[8];
	if((cp != 0) && (cp != thispage) ){
		skip = 1; /* Seek to next cp */
	} 
	else 
	{
		printf("Code page %u found.\n",thispage);
	}
	fontsleft=numfonts;
	while (fontsleft--){
		if (fread(hdr, 6, 1, fp) != 1) {
			perror("Error reading .cp subheader");
			return(0);
		}
		if((hdr[1]==8)&&(!skip)){
			if((height==0)&&(numfonts==1)) return(hdr[0]); /* didn't care, this will do */
			if(hdr[0]==height) return(height);
		}
		fseek(fp,256*hdr[0],SEEK_CUR);/* seek to end of font */
	}
	if(cp == thispage) printf("Multiple sizes, which one?");
	printf("\n");
	return(0);
}

int filesize(int fd){
	/* pass a file descriptor (maybe fp->_fileno, fp is FILE*), get the size. */
	struct stat stbuf;
	if (fstat(fd, &stbuf)) {
		perror("Couldn't stat file");
		exit(1);
	}
	return(stbuf.st_size);
}

int uploadfont(struct kgi_font *font)
{
	int fd;
	
	if ((fd = open("/dev/graphic", 0)) < 0) 
		fd = 0;
	if(ioctl(fd, CHIP_SETDEFFONT, font)) 
	{
		perror("CHIP_SETDEFFONT ioctl error");
		close(fd);
		return(1);
	} else 
		printf("Uploaded font to kernel.\n");
	close(fd);
	return(0);
}
